%{

#include <Windows.h>
#include <stdlib.h>
#include <string.h>

int yylex( void );
int yyparse( void );

LPSTR szIn, szOut;
WORD  wPos;

void pascal ShowMessage( LPSTR );

#define P(x) ShowMessage(x)

%}

%union
{
   LPSTR String;
   long Number;
}

%token <Number> NUMBER
%token <String> STRING
%token <String> IDENTIFIER
%token <Number> NOT
%token _TRUE
%token _FALSE

%left '+' '-' '*'
%right '('

%%

Expr      : NUMBER              { P("NUMBER"); }
          | STRING              { P("STRING"); }
          | FunCall             { P("FUNCTION");}
          | Object
          | Expr '+' Expr       { P("PLUS"); }
          | Expr '-' Expr       { P("SUB"); }
          | Expr '*' Expr       { P("Mult"); }
          | '(' Expr ')'
          ;

FunCall   : IDENTIFIER '(' ')'            { P($1); }
          | IDENTIFIER '(' ParamList ')'  { P($1); }
          ;

ParamList : Expr
          | ParamList ',' Expr
          ;

Object    : IDENTIFIER                               { P($1); }
          | Object '.' IDENTIFIER                    { P($3); }
          | Object '.' IDENTIFIER '(' ')'            { P($3); }
          | Object '.' IDENTIFIER '(' ParamList ')'  { P($3); }
          ;

%%

static YYSTYPE yylval;

int yyerror( LPSTR szMsg )
{
   MessageBox( 0, szMsg, "Error", MB_ICONINFORMATION );
   return 1;
}

WORD ChrAt( BYTE byte, LPSTR szText )
{
   WORD w = 1;

   while( szText[ w - 1 ] && szText[ w - 1 ] != byte )
      w++;

   if( szText[ w - 1 ] == byte )
      return w;
   else
      return 0;
}

int yylex( void )
{
   static BYTE token[ 25 ];
   BYTE b = 0;

   if( wPos == strlen( szIn ) )
      return -1;

   while( szIn[ wPos ] == ' ' )
     wPos++;

   if( szIn[ wPos ] >= '0' && szIn[ wPos ] <= '9' )
   {
      while( szIn[ wPos ] >= '0' && szIn[ wPos ] <= '9' )
         token[ b++ ] = szIn[ wPos++ ];
      token[ b ] = 0;

      yylval.Number = atol( token );
      return NUMBER;
   }

   if( szIn[ wPos ] == '"' )
   {
      token[ b++ ] = '"';
      wPos++;

      while( szIn[ wPos ] != '"' )
      {
         token[ b++ ] = '"';
         wPos++;
      }

      token[ b++ ] = '"';
      token[ b ] = 0;
      wPos++;
      yylval.String = token;

      return STRING;
   }

   if( szIn[ wPos ] >= 'A' && szIn[ wPos ] <= 'z' )
   {
      while( szIn[ wPos ] >= 'A' && szIn[ wPos ] <= 'z' )
         token[ b++ ] = szIn[ wPos++ ];

      token[ b ] = 0;
      yylval.String = token;

      return IDENTIFIER;
   }

   if( ChrAt( szIn[ wPos ], "+-(),.*/" ) )
      return szIn[ wPos++ ];

   return -1;
}

void pascal parse( LPSTR szExpression, LPSTR szBuffer )
{
   szIn  = szExpression;
   szOut = szBuffer;
   wPos  = 0;

   yyparse();
}
